/*
 * Copyright 2010-2012 Luca Garulli (l.garulli--at--orientechnologies.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.orientechnologies.orient.server.replication;

import java.io.IOException;

import com.orientechnologies.common.log.OLogManager;
import com.orientechnologies.orient.core.Orient;
import com.orientechnologies.orient.core.db.ODatabase;
import com.orientechnologies.orient.core.db.ODatabaseComplex;
import com.orientechnologies.orient.core.db.ODatabaseLifecycleListener;
import com.orientechnologies.orient.core.db.record.ORecordOperation;
import com.orientechnologies.orient.core.hook.ORecordHook;
import com.orientechnologies.orient.core.record.ORecord;
import com.orientechnologies.orient.core.record.ORecordInternal;
import com.orientechnologies.orient.core.record.impl.ODocument;
import com.orientechnologies.orient.core.storage.OStorage;
import com.orientechnologies.orient.server.network.protocol.distributed.OReplicationActiveThreadLocal;

/**
 * Record hook implementation. Catches all the relevant events and propagates to the other cluster nodes.
 * 
 * @author Luca Garulli (l.garulli--at--orientechnologies.com)
 */
public class OReplicatorRecordHook implements ORecordHook, ODatabaseLifecycleListener {

	private OReplicator	replicator;

	/**
	 * Auto install itself as lifecycle listener for databases.
	 */
	public OReplicatorRecordHook(final OReplicator iReplicator) {
		replicator = iReplicator;
		Orient.instance().addDbLifecycleListener(this);
	}

	@Override
	public boolean onTrigger(final TYPE iType, final ORecord<?> iRecord) {
		if (!OReplicationActiveThreadLocal.INSTANCE.get())
			// REPLICATED RECORD, AVOID TO PROPAGATE IT AGAIN
			return false;

		if (iRecord instanceof ODocument && replicator.isIgnoredDocumentClass(((ODocument) iRecord).getClassName()))
			// AVOID TO REPLICATE THE CONFLICT
			return false;

		if (!replicator.isReplicated(iRecord.getDatabase().getName()))
			// DB NOT REPLICATED
			return false;

		try {
			switch (iType) {
			case AFTER_CREATE:
				if (checkRecord(iRecord))
					replicator.distributeRequest(new ORecordOperation((ORecordInternal<?>) iRecord, ORecordOperation.CREATED));
				break;

			case AFTER_UPDATE:
				if (checkRecord(iRecord))
					replicator.distributeRequest(new ORecordOperation((ORecordInternal<?>) iRecord, ORecordOperation.UPDATED));
				break;

			case AFTER_DELETE:
				if (checkRecord(iRecord))
					replicator.distributeRequest(new ORecordOperation((ORecordInternal<?>) iRecord, ORecordOperation.DELETED));
				break;
			}
		} catch (IOException e) {
			throw new ODistributedSynchronizationException("Error on distribution of the record to the configured cluster", e);
		}
		return false;
	}

	/**
	 * Install the itself as trigger to catch all the events against records
	 */
	@Override
	public void onOpen(final ODatabase iDatabase) {
		((ODatabaseComplex<?>) iDatabase).registerHook(this);
	}

	/**
	 * Remove itself as trigger to catch all the events against records
	 */
	@Override
	public void onClose(final ODatabase iDatabase) {
		((ODatabaseComplex<?>) iDatabase).unregisterHook(this);
	}

	protected boolean checkRecord(final ORecord<?> iRecord) {
		final int clId = iRecord.getIdentity().getClusterId();

		if (clId == iRecord.getDatabase().getClusterIdByName(OStorage.CLUSTER_INTERNAL_NAME)
				|| clId == iRecord.getDatabase().getClusterIdByName(OStorage.CLUSTER_INDEX_NAME)) {
			OLogManager.instance().warn(this,
					"Changes to the cluster %s.%s cannot be propagated to remote nodes. Assure to align the schema manually",
					iRecord.getDatabase().getName(), OStorage.CLUSTER_INTERNAL_NAME);
			return false;
		}
		return true;
	}
}
